"""Location handler for Crownstone cloud data."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator

if TYPE_CHECKING:
    from crownstone_cloud.cloud import CrownstoneCloud


class Locations:
    """Handler for the locations of a sphere."""

    def __init__(self, cloud: CrownstoneCloud, sphere_id: str) -> None:
        """Initialization."""
        self.cloud = cloud
        self.sphere_id = sphere_id
        self.data: dict[str, Location] = {}

    def __iter__(self) -> Iterator[Location]:
        """Iterate over locations."""
        return iter(self.data.values())

    async def async_update_location_data(self) -> None:
        """
        Get the locations and presence from the cloud.

        This method is a coroutine.
        """
        cloud_data: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "Spheres", "ownedLocations", model_id=self.sphere_id
        )
        # process items
        removed_items: list[str] = []
        new_items: list[str] = []
        for location in cloud_data:
            location_id: str = location["id"]
            exists = self.data.get(location_id)
            # check if the location already exists
            # it is important that we don't throw away existing objects,
            # as they need to remain functional
            if exists:
                # update data
                self.data[location_id].data = location
            else:
                # add new Location
                self.data[location_id] = Location(location)

            # generate list with new id's to check with the existing id's
            new_items.append(location_id)

        # check for removed items
        for location_id in self.data:
            if location_id not in new_items:
                removed_items.append(location_id)

        # remove items from dict
        for location_id in removed_items:
            del self.data[location_id]

    async def async_update_location_presence(self) -> None:
        """
        Get the presentPeople for this Sphere, and get the users in this Location.

        This method is a coroutine.
        """
        presence_data: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "Spheres", "presentPeople", model_id=self.sphere_id
        )
        # reset the presence
        for location in self.data.values():
            location.present_people = []
        # add new presence
        for presence in presence_data:
            for present_location in presence["locations"]:
                for location in self.data.values():
                    if present_location == location.cloud_id:
                        location.present_people.append(presence["userId"])

    def find(self, location_name: str) -> Location | None:
        """Search for a sphere by name and return sphere object if found."""
        for location in self.data.values():
            if location_name == location.name:
                return location

        return None

    def find_by_id(self, location_id: str) -> Location | None:
        """Search for a sphere by id and return sphere object if found."""
        return self.data.get(location_id)


class Location:
    """Represents a Location."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialization."""
        self.data = data
        self.present_people: list[str] = []

    @property
    def name(self) -> str:
        """Return the name of this Location."""
        return str(self.data["name"])

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this Location."""
        return str(self.data["id"])

    @property
    def unique_id(self) -> int:
        """Return the unique id of this Location."""
        return int(self.data["uid"])



"""Crownstone handler for Crownstone cloud data"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator

from crownstone_cloud.const import DIMMING_ABILITY
from crownstone_cloud.exceptions import AbilityError, CrownstoneAbilityError

if TYPE_CHECKING:
    from crownstone_cloud.cloud import CrownstoneCloud


class Crownstones:
    """Handler for the crownstones of a sphere."""

    def __init__(self, cloud: CrownstoneCloud, sphere_id: str) -> None:
        """Initialization."""
        self.cloud = cloud
        self.sphere_id = sphere_id
        self.data: dict[str, Crownstone] = {}

    def __iter__(self) -> Iterator[Crownstone]:
        """Iterate over crownstones."""
        return iter(self.data.values())

    async def async_update_crownstone_data(self) -> None:
        """Get the crownstones data from the cloud."""
        # include abilities and current switch state in the request
        data_filter = {"include": ["currentSwitchState", {"abilities": "properties"}]}
        # request data
        cloud_data: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "Spheres", "ownedStones", data_filter=data_filter, model_id=self.sphere_id
        )
        # process items
        removed_items: list[str] = []
        new_items: list[str] = []
        for crownstone in cloud_data:
            crownstone_id: str = crownstone["id"]
            exists = self.data.get(crownstone_id)
            # check if the crownstone already exists
            # it is important that we don't throw away existing objects,
            # as they need to remain functional
            if exists:
                # update data and update abilities
                self.data[crownstone_id].data = crownstone
            else:
                # add new Crownstone
                self.data[crownstone_id] = Crownstone(self.cloud, crownstone)

            # update the abilities of the Crownstone from the data
            self.data[crownstone_id].update_abilities()

            # generate list with new id's to check with the existing id's
            new_items.append(crownstone_id)

        # check for removed items
        for crownstone_id in self.data:
            if crownstone_id not in new_items:
                removed_items.append(crownstone_id)

        # remove items from dict
        for crownstone_id in removed_items:
            del self.data[crownstone_id]

    def find(self, crownstone_name: str) -> Crownstone | None:
        """Search for a crownstone by name and return crownstone object if found."""
        for crownstone in self.data.values():
            if crownstone_name == crownstone.name:
                return crownstone

        return None

    def find_by_id(self, crownstone_id: str) -> Crownstone | None:
        """Search for a crownstone by id and return crownstone object if found."""
        return self.data.get(crownstone_id)

    def find_by_uid(self, crownstone_uid: int) -> Crownstone | None:
        """Search for a crownstone by uid and return crownstone object if found."""
        for crownstone in self.data.values():
            if crownstone_uid == crownstone.unique_id:
                return crownstone

        return None


class CrownstoneAbility:
    """Represents a Crownstone Ability"""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialization"""
        self.data = data
        self.is_enabled: bool = self.data["enabled"]

    @property
    def type(self) -> str:
        """Return the ability type."""
        return str(self.data["type"])

    @property
    def ability_id(self) -> str:
        """Return the ability id."""
        return str(self.data["id"])

    @property
    def crownstone_id(self) -> str:
        """Return the Crownstone id."""
        return str(self.data["stoneId"])


class Crownstone:
    """Represents a Crownstone"""

    def __init__(self, cloud: CrownstoneCloud, data: dict[str, Any]) -> None:
        """Initialization."""
        self.cloud = cloud
        self.data = data
        self.abilities: dict[str, CrownstoneAbility] = {}
        # Not cloud data, store your own data here (from Crownstone USB)
        self.power_usage = 0
        self.energy_usage = 0

    @property
    def name(self) -> str:
        """Return the name of this Crownstone."""
        return str(self.data["name"])

    @property
    def unique_id(self) -> int:
        """Return the unique_id of this Crownstone."""
        return int(self.data["uid"])

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this Crownstone."""
        return str(self.data["id"])

    @property
    def type(self) -> str:
        """Return the Crownstone type."""
        return str(self.data["type"])

    @property
    def sw_version(self) -> str:
        """Return the firmware version of this Crownstone."""
        return str(self.data["firmwareVersion"])

    @property
    def icon(self) -> str:
        """Return the icon of this Crownstone."""
        return str(self.data["icon"])

    @property
    def state(self) -> int:
        """Return the last reported switch state (0 - 100) of this Crownstone."""
        return int(self.data["currentSwitchState"]["switchState"])

    @state.setter
    def state(self, value: int) -> None:
        """
        Set a new switch state (0 - 100) of this Crownstone.

        Only updates state in cloud, does not send a switch command to the actual Crownstone.
        """
        self.data["currentSwitchState"]["switchState"] = value

    def update_abilities(self) -> None:
        """Add/update the abilities for this Crownstone."""
        for ability in self.data["abilities"]:
            self.abilities[ability["type"]] = CrownstoneAbility(ability)

    async def async_turn_on(self) -> None:
        """
        Turn this crownstone on.

        This method is a coroutine.
        """
        # send a command to the cloud to turn the Crownstone on.
        await self.cloud.request_handler.post(
            "Stones", "switch", model_id=self.cloud_id, json={"type": "TURN_ON"}
        )

    async def async_turn_off(self) -> None:
        """
        Turn this crownstone off.

        This method is a coroutine.
        """
        # send a command to the cloud to turn the Crownstone off.
        await self.cloud.request_handler.post(
            "Stones", "switch", model_id=self.cloud_id, json={"type": "TURN_OFF"}
        )

    async def async_set_brightness(self, brightness: int) -> None:
        """
        Set the brightness of this crownstone, if dimming enabled.

        :param brightness: brightness value between (0 - 100)

        This method is a coroutine.
        """
        # check dimming availability & value, and send a command to the cloud to dim the Crownstone.
        if self.abilities[DIMMING_ABILITY].is_enabled:
            if brightness < 0 or brightness > 100:
                raise ValueError("Enter a value between 0 and 100")

            await self.cloud.request_handler.post(
                "Stones",
                "switch",
                model_id=self.cloud_id,
                json={"type": "PERCENTAGE", "percentage": brightness},
            )
        else:
            raise CrownstoneAbilityError(AbilityError["NOT_ENABLED"])



user_data = {
    "admins": [
        {
            "firstName": "I am",
            "lastName": "Awesome",
            "new": False,
            "uploadLocation": True,
            "uploadSwitchState": True,
            "uploadDeviceDetails": True,
            "email": "example@example.com",
            "emailVerified": True,
            "id": "i_am_awesome_id",
            "createdAt": "date_created",
            "updatedAt": "date_updated"
        }
    ],
    "members": [
        {
            "firstName": "You are",
            "lastName": "Awesome",
            "new": False,
            "uploadLocation": True,
            "uploadSwitchState": True,
            "uploadDeviceDetails": True,
            "email": "example@example.com",
            "emailVerified": True,
            "id": "you_are_awesome_id",
            "createdAt": "date_created",
            "updatedAt": "date_updated"
        }
    ],
    "guests": [
        {
            "firstName": "We are",
            "lastName": "Awesome",
            "new": False,
            "uploadLocation": True,
            "uploadSwitchState": True,
            "uploadDeviceDetails": True,
            "email": "example@example.com",
            "emailVerified": True,
            "id": "we_are_awesome_id",
            "createdAt": "date_created",
            "updatedAt": "date_updated"
        }
    ]
}


"""Top level imports."""
from crownstone_cloud.cloud import CrownstoneCloud
from crownstone_cloud.helpers.aiohttp_client import create_clientsession
from crownstone_cloud.util.runners import run_async

__version__ = "1.4.11"



"""Functions to run async functions from sync context"""
from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from typing import Any


def run_async(coroutine: Coroutine[Any, Any, None]) -> Any:
    """Run an async function from sync context."""
    return asyncio.get_event_loop().run_until_complete(coroutine)



"""Sphere handler for Crownstone cloud data."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator

from crownstone_cloud.cloud_models.crownstones import Crownstones
from crownstone_cloud.cloud_models.locations import Locations
from crownstone_cloud.cloud_models.users import Users

if TYPE_CHECKING:
    from crownstone_cloud.cloud import CrownstoneCloud


class Spheres:
    """Handler for the spheres of the user."""

    def __init__(self, cloud: CrownstoneCloud, user_id: str) -> None:
        """Initialization."""
        self.cloud = cloud
        self.user_id = user_id
        self.data: dict[str, Sphere] = {}

    def __iter__(self) -> Iterator[Sphere]:
        """Iterate over spheres."""
        return iter(self.data.values())

    async def async_update_sphere_data(self) -> None:
        """
        Get the spheres for the user from the cloud.

        This method is a coroutine.
        """
        cloud_data: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "users", "spheres", model_id=self.user_id
        )
        # process items
        removed_items: list[str] = []
        new_items: list[str] = []
        for sphere in cloud_data:
            sphere_id: str = sphere["id"]
            exists = self.data.get(sphere_id)
            # check if the Sphere already exists
            # it is important that we don't throw away existing objects,
            # as they need to remain functional
            if exists:
                # update data
                self.data[sphere_id].data = sphere
            else:
                # add new Sphere
                self.data[sphere_id] = Sphere(self.cloud, sphere, self.user_id)

            # generate list with new id's to check with the existing id's
            new_items.append(sphere_id)

        # check for removed items
        for sphere_id in self.data:
            if sphere_id not in new_items:
                removed_items.append(sphere_id)

        # remove items from dict
        for sphere_id in removed_items:
            del self.data[sphere_id]

    def find(self, sphere_name: str) -> Sphere | None:
        """Search for a sphere by name and return sphere object if found."""
        for sphere in self.data.values():
            if sphere_name == sphere.name:
                return sphere

        return None

    def find_by_id(self, sphere_id: str) -> Sphere | None:
        """Search for a sphere by id and return sphere object if found."""
        return self.data.get(sphere_id)


class Sphere:
    """Represents a Sphere."""

    def __init__(
        self, cloud: CrownstoneCloud, data: dict[str, Any], user_id: str
    ) -> None:
        """Initialization."""
        self.cloud = cloud
        self.data = data
        self.user_id = user_id
        self.crownstones = Crownstones(self.cloud, self.cloud_id)
        self.locations = Locations(self.cloud, self.cloud_id)
        self.users = Users(self.cloud, self.cloud_id)
        self.keys: dict[str, str] = {}
        self.present_people: list[str] = []

    @property
    def name(self) -> str:
        """Returns the name of this Sphere."""
        return str(self.data["name"])

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this Sphere."""
        return str(self.data["id"])

    @property
    def unique_id(self) -> int:
        """Return the unique id of this Sphere."""
        return int(self.data["uid"])

    async def async_get_keys(self) -> dict[str, str]:
        """
        Get the user keys for this sphere, that can be used for BLE (optional).

        This method is a coroutine.
        """
        # get & reformat keys.
        keys: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "users", "keysV2", model_id=self.user_id
        )
        for key_set in keys:
            if key_set["sphereId"] == self.cloud_id:
                for key_type in key_set["sphereKeys"]:
                    self.keys[key_type["keyType"]] = key_type["key"]

        return self.keys

    async def async_update_sphere_presence(self) -> None:
        """
        Replaces the current presence with that of the cloud.

        This method is a coroutine.
        """
        # get presence and create a list with user id's who are in the sphere.
        self.present_people = []
        presence_data: list[dict[str, Any]] = await self.cloud.request_handler.get(
            "Spheres", "presentPeople", model_id=self.cloud_id
        )
        for user in presence_data:
            self.present_people.append(user["userId"])



"""
This is an example how to switch a crownstone using the crownstone python lib cloud.

Last update by Ricardo Steijn on 31-10-2020
"""
from crownstone_cloud import CrownstoneCloud, run_async
import logging

# Enable logging.
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)

# Initialize cloud.
cloud = CrownstoneCloud('email', 'password')
# Use 'run_async' to run async functions in sync context.
# Login & synchronize all cloud data.
run_async(cloud.async_initialize())

# Get a crownstone by name and turn it on.
crownstone_coffee_machine = cloud.get_crownstone('Coffee machine')
run_async(crownstone_coffee_machine.async_turn_on())

# Close the session after we are done.
run_async(cloud.async_close_session())



"""Functions to get or create an aiohttp clientsession."""
import ssl
from typing import Any

import aiohttp
import certifi


def create_clientsession(**kwargs: Any) -> aiohttp.ClientSession:
    """Create a new aiohttp clientsession."""
    connector = get_connector()

    clientsession = aiohttp.ClientSession(
        connector=connector,
        **kwargs,
    )

    return clientsession


def get_connector() -> aiohttp.BaseConnector:
    """Return the connector for aiohttp."""

    def client_context() -> ssl.SSLContext:
        """Return an SSL context for making requests."""
        context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
        )
        return context

    connector = aiohttp.TCPConnector(enable_cleanup_closed=True, ssl=client_context())

    return connector



from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch

from crownstone_cloud.helpers.requests import RequestHandler
from crownstone_cloud import CrownstoneCloud
from tests.mocked_replies.errors import (
    auth_error,
    access_token_expired,
    not_verified
)
from crownstone_cloud.exceptions import CrownstoneAuthenticationError


class TestCrownstoneCloud(IsolatedAsyncioTestCase):
    """Test the request handler"""

    async def test_exceptions(self):
        cloud = CrownstoneCloud('email', 'password')

        # mock login with errors
        with self.assertRaises(CrownstoneAuthenticationError) as login_err:
            await cloud.request_handler.raise_on_error(auth_error)

        assert login_err.exception.type == 'LOGIN_FAILED'

        with self.assertRaises(CrownstoneAuthenticationError) as not_verified_err:
            await cloud.request_handler.raise_on_error(not_verified)

        assert not_verified_err.exception.type == 'LOGIN_FAILED_EMAIL_NOT_VERIFIED'

        await cloud.async_close_session()

    async def test_refresh_token(self):
        cloud = CrownstoneCloud('email', 'password')

        # mock access_token expired
        with patch.object(RequestHandler, 'request_login') as refresh_mock:
            result = await cloud.request_handler.raise_on_error(access_token_expired)

        assert result is True
        refresh_mock.assert_called()
        refresh_mock.assert_awaited()

        await cloud.async_close_session()



"""Exceptions for crownstone cloud connection cloud"""
from __future__ import annotations

AuthError = {
    "LOGIN_FAILED": "Wrong email or password provided",
    "USERNAME_EMAIL_REQUIRED": "Email or password not provided",
    "LOGIN_FAILED_EMAIL_NOT_VERIFIED": "Email has not been verified, please do that first",
}

AbilityError = {
    "NOT_ENABLED": "This ability is not enabled. Use the Crownstone App to enable it.",
}


class CrownstoneException(Exception):
    """Raised when authentication with API ended in error"""

    def __init__(self, exception_type: str, message: str | None = None):
        super().__init__(exception_type, message)
        self.type = exception_type
        self.message = message


class CrownstoneAuthenticationError(CrownstoneException):
    """Raised when authentication with API ended in error"""


class CrownstoneUnknownError(Exception):
    """Raised when the error is not known / no data"""


class CrownstoneConnectionError(Exception):
    """Raised when a connection to the Crownstone Cloud can't be made."""


class CrownstoneAbilityError(CrownstoneException):
    """Raised when when e.g. dimming command is requested when it is not enabled."""


class CrownstoneNotFoundError(Exception):
    """Raised when searching for a Crownstone object, but it is not found."""



"""Contains functions that convert a parameter into something else."""
from __future__ import annotations

import hashlib
import json
from typing import Any
from urllib.parse import quote


def password_to_hash(password: str) -> str | None:
    """Generate a sha1 password from string."""
    if password is None:
        return None
    pw_hash = hashlib.sha1(password.encode("utf-8"))
    return pw_hash.hexdigest()


def quote_json(_json: dict[str, Any]) -> str:
    """Convert JSON to HTML quote."""
    stringified = json.dumps(_json)
    return quote(stringified)



"""Main class for the Crownstone cloud cloud."""
from __future__ import annotations

import asyncio
import logging

import aiohttp

from crownstone_cloud.cloud_models.crownstones import Crownstone
from crownstone_cloud.cloud_models.spheres import Spheres
from crownstone_cloud.exceptions import CrownstoneNotFoundError
from crownstone_cloud.helpers.conversion import password_to_hash
from crownstone_cloud.helpers.requests import RequestHandler

_LOGGER = logging.getLogger(__name__)


class CrownstoneCloud:
    """Create a Crownstone cloud instance."""

    cloud_data: Spheres
    access_token: str

    def __init__(
        self,
        email: str,
        password: str,
        clientsession: aiohttp.ClientSession | None = None,
    ) -> None:
        self.request_handler = RequestHandler(self, clientsession)
        self.login_data = {"email": email, "password": password_to_hash(password)}

    async def async_initialize(self) -> None:
        """
        Login to Crownstone API & synchronize all cloud data.

        This method is a coroutine.
        """
        # Login
        login_response = await self.request_handler.request_login(self.login_data)

        # Save access token & create cloud data object
        self.access_token = login_response["id"]
        self.cloud_data = Spheres(self, login_response["userId"])
        _LOGGER.debug("Login to Crownstone Cloud successful")

        # Synchronize data
        await self.async_synchronize()

    async def async_synchronize(self) -> None:
        """
        Sync all data from cloud.

        This method is a coroutine.
        """
        _LOGGER.debug("Initiating all cloud data")
        # get the sphere data for this user_id
        await self.cloud_data.async_update_sphere_data()

        # get the data from the sphere attributes
        for sphere in self.cloud_data:
            await asyncio.gather(
                sphere.async_update_sphere_presence(),
                sphere.crownstones.async_update_crownstone_data(),
                sphere.locations.async_update_location_data(),
                sphere.locations.async_update_location_presence(),
                sphere.users.async_update_user_data(),
            )
        _LOGGER.debug("Cloud data successfully initialized")

    def get_crownstone(
        self, crownstone_name: str, sphere_id: str | None = None
    ) -> Crownstone:
        """
        Get a Crownstone object by providing the name and optionally a sphere id.

        :param crownstone_name: Name of the Crownstone.
        :param sphere_id: Sphere id that should match.
        :return: Crownstone object.
        """
        for sphere in self.cloud_data:
            if sphere_id is not None:
                if sphere.cloud_id != sphere_id:
                    continue

            for crownstone in sphere.crownstones:
                if crownstone.name == crownstone_name:
                    return crownstone

        raise CrownstoneNotFoundError from None

    def get_crownstone_by_id(
        self, crownstone_id: str, sphere_id: str | None = None
    ) -> Crownstone:
        """
        Get a Crownstone object by providing the id and optionally a sphere id.

        :param crownstone_id: The cloud id of the Crownstone.
        :param sphere_id: Sphere id that should match.
        :return: Crownstone object.
        """
        for sphere in self.cloud_data:
            if sphere_id is not None:
                if sphere.cloud_id != sphere_id:
                    continue

            for crownstone in sphere.crownstones:
                if crownstone.cloud_id == crownstone_id:
                    return crownstone

        raise CrownstoneNotFoundError from None

    def get_crownstone_by_uid(
        self, crownstone_uid: int, sphere_id: str | None = None
    ) -> Crownstone:
        """
        Get a Crownstone object by providing the uid and optionally a sphere id.

        :param crownstone_uid: The unique id of the Crownstone.
        :param sphere_id: Sphere id that should match.
        :return: Crownstone object.
        """
        for sphere in self.cloud_data:
            if sphere_id is not None:
                if sphere.cloud_id != sphere_id:
                    continue

            for crownstone in sphere.crownstones:
                if crownstone.unique_id == crownstone_uid:
                    return crownstone

        raise CrownstoneNotFoundError from None

    async def async_close_session(self) -> None:
        """
        Close the aiohttp clientsession after all requests are done.

        The session should always be closed when the program ends.
        When there's an external clientsession in use, DON'T use this method.

        This method is a coroutine.
        """
        await self.request_handler.client_session.close()



"""Mock replies from the cloud for Crownstone data."""

crownstone_data = [
    {
        "name": "my_awesome_crownstone",
        "address": "address",
        "type": "PLUG",
        "dimmingEnabled": "false",
        "major": 69,
        "minor": 69,
        "uid": 1,
        "icon": "fiCS1-desk-lamp-2",
        "tapToToggle": "true",
        "firmwareVersion": "5.0.1",
        "bootloaderVersion": "2.1.0",
        "hardwareVersion": "101020100000000000000000000QFAAB0",
        "hidden": "false",
        "locked": "false",
        "id": "my_awesome_crownstone_id",
        "locationId": "location_id",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated",
        "currentSwitchStateId": "switch_state_id",
        "currentSwitchState": {
            "timestamp": "date_updated",
            "switchState": 0
        },
        "abilities": [
            {
                "type": "dimming",
                "enabled": True,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_1",
                "stoneId": "my_awesome_crownstone_id",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": [
                    {
                        "type": "softOnSpeed",
                        "value": "5",
                        "id": "my_awesome_property_id_1",
                        "abilityId": "my_awesome_ability_id_1",
                        "sphereId": "my_awesome_sphere_id",
                        "createdAt": "date_created",
                        "updatedAt": "date_updated"
                    }
                ]
            },
            {
                "type": "switchcraft",
                "enabled": False,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_2",
                "stoneId": "my_awesome_crownstone_id",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": []
            },
            {
                "type": "tapToToggle",
                "enabled": False,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_3",
                "stoneId": "my_awesome_crownstone_id",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": []
            }
        ]
    },
    {
        "name": "my_awesome_crownstone_2",
        "address": "address",
        "type": "PLUG",
        "dimmingEnabled": "false",
        "major": 69,
        "minor": 69,
        "uid": 1,
        "icon": "fiCS1-desk-lamp-2",
        "tapToToggle": "true",
        "firmwareVersion": "5.0.1",
        "bootloaderVersion": "2.1.0",
        "hardwareVersion": "101020100000000000000000000QFAAB0",
        "hidden": "false",
        "locked": "false",
        "id": "my_awesome_crownstone_id_2",
        "locationId": "location_id",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated",
        "currentSwitchStateId": "switch_state_id",
        "currentSwitchState": {
            "timestamp": "date_updated",
            "switchState": 0
        },
        "abilities": [
            {
                "type": "dimming",
                "enabled": False,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_4",
                "stoneId": "my_awesome_crownstone_id_2",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": []
            },
            {
                "type": "switchcraft",
                "enabled": False,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_5",
                "stoneId": "my_awesome_crownstone_id_2",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": []
            },
            {
                "type": "tapToToggle",
                "enabled": True,
                "syncedToCrownstone": True,
                "id": "my_awesome_ability_id_6",
                "stoneId": "my_awesome_crownstone_id_2",
                "sphereId": "my_awesome_sphere_id",
                "createdAt": "date_created",
                "updatedAt": "date_updated",
                "properties": [
                    {
                        "type": "rssiOffset",
                        "value": "-10",
                        "id": "my_awesome_property_id_2",
                        "abilityId": "my_awesome_ability_id_6",
                        "sphereId": "my_awesome_sphere_id",
                        "createdAt": "date_created",
                        "updatedAt": "date_updated"
                    }
                ]
            }
        ]
    },
]



"""Handler for requests to the Crownstone cloud"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from aiohttp import ClientSession, ContentTypeError

from crownstone_cloud.const import DEFAULT_CLOUD_ADDR, CLOUD_LOGIN_SUFFIX
from crownstone_cloud.exceptions import (
    AuthError,
    CrownstoneAuthenticationError,
    CrownstoneConnectionError,
    CrownstoneUnknownError,
)
from crownstone_cloud.helpers.aiohttp_client import create_clientsession
from crownstone_cloud.helpers.conversion import quote_json

if TYPE_CHECKING:
    from crownstone_cloud.cloud import CrownstoneCloud

_LOGGER = logging.getLogger(__name__)


class RequestHandler:
    """Handles requests to the Crownstone cloud."""

    def __init__(
        self, cloud: CrownstoneCloud, clientsession: ClientSession | None = None
    ) -> None:
        self.cloud = cloud
        self.client_session = clientsession or create_clientsession()

    async def request_login(self, login_data: dict[str, Any]) -> Any:
        """Request a login to the Crownstone Cloud API."""
        response = await self.request("post", f"{DEFAULT_CLOUD_ADDR}{CLOUD_LOGIN_SUFFIX}", login_data)

        return response

    async def post(
        self,
        model: str,
        endpoint: str,
        model_id: str | None = None,
        json: dict[str, Any] | None = None,
    ) -> Any:
        """
        Post request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, presentPeople.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :param json: Dictionary with the data that should be posted.
        :return: Dictionary with the response from the cloud.
        """
        if model_id:
            url = f"{DEFAULT_CLOUD_ADDR}/{model}/{model_id}/{endpoint}?access_token={self.cloud.access_token}"
        else:
            url = f"{DEFAULT_CLOUD_ADDR}/{model}{endpoint}?access_token={self.cloud.access_token}"

        return await self.request("post", url, json)

    async def get(
        self,
        model: str,
        endpoint: str,
        data_filter: dict[str, Any] | None = None,
        model_id: str | None = None,
    ) -> Any:
        """
        Get request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, preUnionentPeople.
        :param filter: filter output or add extra data to output.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :return: Dictionary with the response from the cloud.
        """
        if data_filter and model_id:
            url = (
                f"{DEFAULT_CLOUD_ADDR}/{model}/{model_id}/{endpoint}?filter={quote_json(data_filter)}&access_token="
                f"{self.cloud.access_token}"
            )
        elif model_id and not data_filter:
            url = f"{DEFAULT_CLOUD_ADDR}/{model}/{model_id}/{endpoint}?access_token={self.cloud.access_token}"
        else:
            url = f"{DEFAULT_CLOUD_ADDR}/{model}{endpoint}?access_token={self.cloud.access_token}"

        return await self.request("get", url)

    async def put(
        self, model: str, endpoint: str, model_id: str, command: str, value: Any
    ) -> Any:
        """
        Put request

        :param model: model type. users, spheres, stones, locations, devices.
        :param endpoint: endpoints. e.g. spheres, keys, presentPeople.
        :param model_id: required id for the endpoint. e.g. userId for users, sphereId for spheres.
        :param command: used for command requests. e.g. 'switchState'.
        :param value: the value to be put for the command. e.g 'switchState', 1
        :return: Dictionary with the response from the cloud.
        """
        url = (
            f"{DEFAULT_CLOUD_ADDR}/{model}/{model_id}/{endpoint}?{command}={str(value)}&access_token="
            f"{self.cloud.access_token}"
        )

        return await self.request("put", url)

    async def request(
        self, method: str, url: str, json: dict[str, Any] | None = None
    ) -> Any:
        """Make request and check data for errors."""
        async with self.client_session.request(method, url, json=json) as result:
            try:
                data: Any = await result.json()
            except ContentTypeError as err:
                # when the cloud is unavailable,
                # a payload can be received that can't be converted to a dictionary.
                raise CrownstoneConnectionError(
                    "Error connecting to the Crownstone Cloud."
                ) from err
            refresh = await self.raise_on_error(data)
            if refresh:
                new_url = url.replace(
                    url.split("access_token=", 1)[1], self.cloud.access_token
                )
                await self.request(method, new_url, json=json)
            return data

    async def raise_on_error(self, data: Any) -> bool:
        """Check for error messages and raise the correct exception."""
        if isinstance(data, dict) and "error" in data:
            error: dict[str, Any] = data["error"]
            if "code" in error:
                error_type = error["code"]
                try:
                    if error_type in ("INVALID_TOKEN", "AUTHORIZATION_REQUIRED"):
                        # Login using existing data
                        response = await self.request_login(self.cloud.login_data)
                        self.cloud.access_token = response["id"]
                        return True  # re-run the request

                    for err_type, message in AuthError.items():
                        if err_type == error_type:
                            raise CrownstoneAuthenticationError(err_type, message)

                except ValueError as err:
                    raise CrownstoneUnknownError("Unknown error occurred.") from err
            else:
                _LOGGER.error(error["message"])

        return False



"""User handler for Crownstone cloud data."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator

if TYPE_CHECKING:
    from crownstone_cloud.cloud import CrownstoneCloud


class Users:
    """Handler for the users in a sphere."""

    def __init__(self, cloud: CrownstoneCloud, sphere_id: str) -> None:
        """Initialization."""
        self.cloud = cloud
        self.sphere_id = sphere_id
        self.data: dict[str, User] = {}

    def __iter__(self) -> Iterator[User]:
        """Iterate over users."""
        return iter(self.data.values())

    async def async_update_user_data(self) -> None:
        """
        Get the users for the sphere from the cloud.

        This method is a coroutine.
        """
        cloud_data: dict[
            str, list[dict[str, Any]]
        ] = await self.cloud.request_handler.get(
            "Spheres", "users", model_id=self.sphere_id
        )
        # process items
        removed_items: list[str] = []
        new_items: list[str] = []
        for role, users in cloud_data.items():
            for user in users:
                user_id = user["id"]
                exists = self.data.get(user_id)
                # check if the User already exists
                # it is important that we don't throw away existing objects,
                # as they need to remain functional
                if exists:
                    # update data
                    self.data[user_id].data = user
                else:
                    # add new User + their role
                    self.data[user_id] = User(user, role)

                # generate list with new id's to check with the existing id's
                new_items.append(user_id)

        # check for removed items
        for user_id in self.data:
            if user_id not in new_items:
                removed_items.append(user_id)

        # remove items from dict
        for user_id in removed_items:
            del self.data[user_id]

    def find_by_first_name(self, first_name: str) -> list[User]:
        """Search for a user by first name and return a list with the users found."""
        found_users: list[User] = []
        for user in self.data.values():
            if first_name == user.first_name:
                found_users.append(user)

        return found_users

    def find_by_last_name(self, last_name: str) -> list[User]:
        """Search for a user by last name and return a list with the users found."""
        found_users: list[User] = []
        for user in self.data.values():
            if last_name == user.last_name:
                found_users.append(user)

        return found_users

    def find_by_id(self, user_id: str) -> User | None:
        """Search for a user by id and return crownstone object if found."""
        return self.data.get(user_id)


class User:
    """Represents a user in a sphere."""

    def __init__(self, data: dict[str, Any], role: str) -> None:
        """Initialization."""
        self.data = data
        self.role = role

    @property
    def first_name(self) -> str:
        """Return the first name of this User."""
        return str(self.data["firstName"])

    @property
    def last_name(self) -> str:
        """Return the last name of this User."""
        return str(self.data["lastName"])

    @property
    def email(self) -> str:
        """Return the email of this User."""
        return str(self.data["email"])

    @property
    def cloud_id(self) -> str:
        """Return the cloud id of this User."""
        return str(self.data["id"])

    @property
    def email_verified(self) -> bool:
        """Return whether the user has verified email."""
        return bool(self.data["emailVerified"])



location_data_init = [
    {
        "name": "my_awesome_location_1",
        "uid": 1,
        "icon": "c1-tvSetup2",
        "id": "my_awesome_location_id_1",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    },
    {
        "name": "my_awesome_location_2",
        "uid": 2,
        "icon": "c1-foodWine",
        "id": "my_awesome_location_id_2",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    },
    {
        "name": "my_awesome_location_3",
        "uid": 3,
        "icon": "c1-bed",
        "id": "my_awesome_location_id_3",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    }
]

location_data_removed = [
    {
        "name": "my_awesome_location_1",
        "uid": 1,
        "icon": "c1-tvSetup2",
        "id": "my_awesome_location_id_1",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    },
    {
        "name": "my_awesome_location_3",
        "uid": 3,
        "icon": "c1-bed",
        "id": "my_awesome_location_id_3",
        "sphereId": "my_awesome_sphere_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    }
]

presence_data = [
    {
        "userId": "i_am_awesome_id",
        "locations": [
            "my_awesome_location_id_1",
            "my_awesome_location_id_2",
        ]
    },
    {
        "userId": "you_are_awesome_id",
        "locations": [
            "my_awesome_location_id_3",
        ]
    },
    {
        "userId": "we_are_awesome_id",
        "locations": [
            "my_awesome_location_id_2",
        ]
    }
]


"""
This is an example how to switch a crownstone using the crownstone python lib cloud.
Using this library in async context is the recommended way.

Last update by Ricardo Steijn on 20-09-2021
"""
from crownstone_cloud import CrownstoneCloud, create_clientsession
import logging
import asyncio

# Enable logging.
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)


async def main():
    # Every instance creates it's own websession for easy accessibility, however using 1 websession is recommended.
    # Create your websession like so:
    websession = create_clientsession()
    # Initialize cloud.
    cloud_user_1 = CrownstoneCloud('email_user_1', 'password_user_1', websession)
    # Login to the Crownstone Cloud and synchronize all cloud data.
    await cloud_user_1.async_initialize()

    # Get a crownstone by name that can dim, and put it on 20% brightness for user 1
    crownstone_lamp = cloud_user_1.get_crownstone('Lamp')
    await crownstone_lamp.async_set_brightness(20)

    # Login & synchronize data for an other account.
    cloud_user_2 = CrownstoneCloud('email_user_2', 'password_user_2', websession)
    await cloud_user_2.async_initialize()

    # Get a crownstone by name and turn it on for user 2.
    crownstone_tv = cloud_user_2.get_crownstone('TV')
    await crownstone_tv.async_turn_on()

    # If you want to update specific data you can get the cloud data object for your user.
    # This object has all the cloud data for your user saved in it, which was synced with async_initialize()
    # Parts of the data can also be synced individually without touching the other data.
    # To sync all data at once, use async_synchronize() instead.
    my_sphere = cloud_user_1.cloud_data.find("my_sphere_name")
    # request to sync only the locations with the cloud
    await my_sphere.locations.async_update_location_data()
    # get the keys for this sphere so you can use them with the Crownstone BLE python library
    sphere_keys = await my_sphere.async_get_keys()

    # Close the aiohttp clientsession after we are done.
    await websession.close()

asyncio.run(main())



"""Constants for the Crownstone cloud library."""
from typing import Final

DEFAULT_CLOUD_ADDR: Final = "https://cloud.ownstone.org/api"
CLOUD_LOGIN_SUFFIX: Final = "/users/login"

# Crownstone abilities
DIMMING_ABILITY: Final = "dimming"
SWITCHCRAFT_ABILITY: Final = "switchcraft"
TAP_TO_TOGGLE_ABILITY: Final = "tapToToggle"



"""Setup for Crownstone Cloud library."""
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='crownstone-cloud',
    version="1.4.11",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/Crownstone-Community/crownstone-lib-python-cloud',
    author='Crownstone B.V.',
    packages=find_packages(exclude=['examples', 'tests']),
    install_requires=list(package.strip() for package in open('requirements.txt')),
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
    python_requires='>=3.8',
)



"""Tests for the Crownstone Cloud library."""
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch

import aiohttp

from crownstone_cloud import CrownstoneCloud
from crownstone_cloud.helpers.requests import RequestHandler
from crownstone_cloud.exceptions import CrownstoneAbilityError, AbilityError
from crownstone_cloud.cloud_models.spheres import Spheres
from crownstone_cloud.const import (
    DIMMING_ABILITY,
    TAP_TO_TOGGLE_ABILITY
)

from tests.mocked_replies.login_data import login_data
from tests.mocked_replies.sphere_data import (
    sphere_data,
    key_data,
    expected_key_data,
)
from tests.mocked_replies.crownstone_data import (
    crownstone_data,
)
from tests.mocked_replies.user_data import user_data
from tests.mocked_replies.location_data import (
    location_data_init,
    location_data_removed,
    presence_data
)


class TestCrownstoneCloud(IsolatedAsyncioTestCase):
    """Test the main class"""

    async def test_init(self):
        """Initialize test case."""
        cloud = CrownstoneCloud('email', 'password')
        assert isinstance(cloud.request_handler.client_session, aiohttp.ClientSession)

        await cloud.async_close_session()

    @patch.object(RequestHandler, 'request_login')
    @patch.object(CrownstoneCloud, 'async_synchronize')
    async def test_initialize(self, mock_sync, mock_request):
        """Test fetching the cloud data."""
        cloud = CrownstoneCloud('email', 'password')
        # patch the result of login request
        mock_request.return_value = login_data
        await cloud.async_initialize()
        assert cloud.access_token == 'my_access_token'
        assert cloud.cloud_data.user_id == 'user_id'
        assert mock_sync.call_count == 1

        await cloud.async_close_session()

    @patch.object(RequestHandler, 'get')
    async def test_data_structure(self, mock_request):
        """Test if data structure is correctly initialized with the cloud data."""
        cloud = CrownstoneCloud('email', 'password')
        # create fake instance
        cloud.spheres = Spheres(cloud, 'user_id')
        # add fake sphere data for test
        mock_request.return_value = sphere_data
        await cloud.spheres.async_update_sphere_data()
        mock_request.assert_awaited()

        # test getting a sphere by id and name
        sphere = cloud.spheres.find_by_id('my_awesome_sphere_id_2')
        assert sphere.cloud_id == 'my_awesome_sphere_id_2'
        sphere2 = cloud.spheres.find('my_awesome_sphere')
        assert sphere2.name == 'my_awesome_sphere'

        # test getting keys
        mock_request.return_value = key_data
        keys = await sphere.async_get_keys()
        mock_request.assert_awaited()
        assert keys == expected_key_data

        # add fake crownstone data for test
        mock_request.return_value = crownstone_data
        await sphere.crownstones.async_update_crownstone_data()
        mock_request.assert_awaited()

        # add fake user data for test
        mock_request.return_value = user_data
        await sphere.users.async_update_user_data()
        mock_request.assert_awaited()

        # add fake location data for test
        mock_request.return_value = location_data_init
        await sphere.locations.async_update_location_data()
        mock_request.assert_awaited()
        # presence
        mock_request.return_value = presence_data
        await sphere.locations.async_update_location_presence()
        mock_request.assert_awaited()

        # test getting crownstone, users, locations by id & name
        crownstone = sphere.crownstones.find('my_awesome_crownstone')
        crownstone_by_id = sphere.crownstones.find_by_id('my_awesome_crownstone_id_2')
        assert crownstone.name == 'my_awesome_crownstone'
        assert crownstone_by_id.cloud_id == 'my_awesome_crownstone_id_2'
        # test state
        assert crownstone.state == 0
        # test abilities
        assert crownstone_by_id.abilities.get(DIMMING_ABILITY).is_enabled is False
        assert crownstone_by_id.abilities.get(TAP_TO_TOGGLE_ABILITY).is_enabled is True

        location = sphere.locations.find('my_awesome_location_1')
        location_by_id = sphere.locations.find_by_id('my_awesome_location_id_3')
        assert location.cloud_id == 'my_awesome_location_id_1'
        assert location_by_id.name == 'my_awesome_location_3'
        # test presence
        assert len(location.present_people) == 1
        assert len(location_by_id.present_people) == 1

        mock_request.return_value = location_data_removed
        await sphere.locations.async_update_location_data()
        # test when the location data gets updated
        assert 'my_awesome_location_id_2' not in sphere.locations.data
        assert 'my_awesome_location_id_3' in sphere.locations.data

        mock_request.return_value = location_data_init
        await sphere.locations.async_update_location_data()
        # test when the presence data gets updated
        assert 'my_awesome_location_id_2' in sphere.locations.data

        user_first = sphere.users.find_by_first_name('I am')
        user_last = sphere.users.find_by_last_name('Awesome')
        user_by_id = sphere.users.find_by_id('we_are_awesome_id')
        assert 'You are' and 'We are' not in user_first
        assert len(user_last) == 3
        assert user_by_id.first_name == 'We are'
        assert user_by_id.last_name == 'Awesome'

        # test setting brightness of a crownstone
        with patch.object(RequestHandler, 'put') as brightness_mock:
            # test if it doesn't run if dimming not enabled
            with self.assertRaises(CrownstoneAbilityError) as ability_err:
                await crownstone_by_id.async_set_brightness(50)

            assert ability_err.exception.type == AbilityError['NOT_ENABLED']
            brightness_mock.assert_not_called()
            # test error when wrong value is given
            with self.assertRaises(ValueError):
                await crownstone.async_set_brightness(101)

        await cloud.async_close_session()



auth_error = {
    "error": {
        "statusCode": 401,
        "name": "Authentication failed",
        "message": "Authentication fail",
        "code": "LOGIN_FAILED",
    }
}

not_verified = {
    "error": {
        "statusCode": 401,
        "name": "Authentication failed",
        "message": "Authentication fail",
        "code": "LOGIN_FAILED_EMAIL_NOT_VERIFIED",
    }
}

access_token_expired = {
    "error": {
        "statusCode": 401,
        "name": "Authentication failed",
        "message": "Authentication fail",
        "code": "INVALID_TOKEN",
    }
}


sphere_data = [
    {
        "name": "my_awesome_sphere",
        "uid": 69,
        "uuid": "uuid",
        "meshAccessAddress": "mesh_access_address",
        "aiName": "smart_ai",
        "aiSex": "null",
        "exitDelay": 600,
        "gpsLocation": {
            "lat": 69,
            "lng": 69
        },
        "id": "my_awesome_sphere_id",
        "ownerId": "owner_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    },
    {
        "name": "my_awesome_sphere_2",
        "uid": 42,
        "uuid": "uuid",
        "meshAccessAddress": "mesh_access_address",
        "aiName": "super_smart_ai",
        "aiSex": "null",
        "exitDelay": 600,
        "gpsLocation": {
            "lat": 42,
            "lng": 42
        },
        "id": "my_awesome_sphere_id_2",
        "ownerId": "owner_id",
        "createdAt": "date_created",
        "updatedAt": "date_updated"
    }
]

key_data = [
  {
    "sphereId": "my_awesome_sphere_id_2",
    "sphereKeys": [
      {
        "id": "admin_id",
        "keyType": "ADMIN_KEY",
        "key": "admin_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "member_id",
        "keyType": "MEMBER_KEY",
        "key": "member_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "basic_id",
        "keyType": "BASIC_KEY",
        "key": "basic_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "localization_id",
        "keyType": "LOCALIZATION_KEY",
        "key": "localization_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "service_data_id",
        "keyType": "SERVICE_DATA_KEY",
        "key": "service_data_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "mesh_application_id",
        "keyType": "MESH_APPLICATION_KEY",
        "key": "mesh_application_key",
        "ttl": 0,
        "createdAt": "date_created"
      },
      {
        "id": "mesh_network_id",
        "keyType": "MESH_NETWORK_KEY",
        "key": "mesh_network_key",
        "ttl": 0,
        "createdAt": "date_created"
      }
    ]
  }
]

expected_key_data = {
    'ADMIN_KEY': 'admin_key',
    'MEMBER_KEY': 'member_key',
    'BASIC_KEY': 'basic_key',
    'LOCALIZATION_KEY': 'localization_key',
    'SERVICE_DATA_KEY': 'service_data_key',
    'MESH_APPLICATION_KEY': 'mesh_application_key',
    'MESH_NETWORK_KEY': 'mesh_network_key'
}
